function alpha = atan2d( y, x )
% -- Function File: atan2d( y, x )
%     Variant of the trigonometric atan function. Calculates the angle between
%     the positive x-axis and the point given by the coordinates x and y. The
%     angle is positive for counter-clockwise angles (upper half-plane, y > 0),
%     and negative for clockwise angles (lower half-plane, y < 0).
%     The output is in the range -180 < atan2d( y, x ) <= 180. atan2d( 0, 0 ) is
%     undefined but returns 0 like other implementations.
%     For better perfomance use the built-in function atan2() * 180/pi.
%
%     Arnold Neugebauer

%little test:
%x = -10:0.1:10;
%y = -10:0.1:10;
%for x_idx = 1:length(x)
%   for y_idx = 1:length(y)
%      alpha  = atan2(  y(y_idx), x(x_idx) ) * 180/pi;
%      alpha1 = atan2d( y(y_idx), x(x_idx) );
%      alpha_error( y_idx, x_idx ) = abs(alpha1 - alpha);
%   endfor
%endfor
%sum( sum(alpha_error) )
%gives something very small like  2.4710e-010

if( x == 0 )
   if( y > 0 )
      alpha = 90;
   elseif( y < 0 )
      alpha = -90;
   else
      alpha = 0; %angle is undefined per definition, return 0 anyway
   endif
else
   alpha = atand( y / x );
   if( x < 0 )
      if( y < 0 )
         alpha -= 180;
      else
         %y=0 is 180 (not -180) per definition
         alpha += 180;
      endif
   endif
endif

endfunction
