function draw_atan_error()
% -- Function File: draw_atan_error()
%     Plots the noise sensitivity of pitch and roll calculation based on the
%     atan function.
%
%     Arnold Neugebauer

acc_strength = 18;
noise_level = acc_strength * 0.1; %10 percent

roll = 0:2:80;
pitch = 0:2:80;

for r_idx = 1:length(roll)
   for p_idx = 1:length(pitch)
      %determine error free sensor outputs
      accx = - acc_strength * sind( pitch(p_idx) );
      accy =   acc_strength * sind( roll(r_idx) ) * cosd( pitch(p_idx) );
      accz =   acc_strength * cosd( roll(r_idx) ) * cosd( pitch(p_idx) );

      %put noise on 'em
      accx = accx + (rand(1)-0.5) * noise_level;
      accy = accy + (rand(1)-0.5) * noise_level;
      accz = accz + (rand(1)-0.5) * noise_level;

      %calculate angles affected by noise
      roll1  =   atand( accy / accz );               %x
      pitch1 = - atand( accx / accz * cosd(roll1) ); %y

      roll_error(p_idx, r_idx)  = abs(roll1 - roll(r_idx));   %x
      pitch_error(p_idx, r_idx) = abs(pitch1 - pitch(p_idx)); %y
   endfor
endfor

subplot( 2, 1, 1 )
mesh( roll, pitch, roll_error );
axis( [0, 80, 0, 80, 0, 45] )
title( "Roll Error Arc Tangent" )
xlabel( "roll" )
ylabel( "pitch" )

subplot( 2, 1, 2 )
mesh( roll, pitch, pitch_error );
axis( [0, 80, 0, 80, 0, 45] )
title( "Pitch Error Arc Tangent" )
xlabel( "roll" )
ylabel( "pitch" )
