%part2.m
% -- Script File: part2
%     The files part1.m to part4.m are both, the text of a tutorial as well as
%     the tutorials Octave scripts. You may just read through them though it is
%     quite handy to see the math actually executed. Walk your way through the
%     parts. Part 1 starts with the Octave setup.
%
%     Arnold Neugebauer

page_screen_output( 0 )     %don't use pager
subplot( 1, 1, 1 )          %set to standard single plot mode

%PART 2
%======

%Basics
%------
%To display results we use vectors a lot. A vector in this tutorial is represented
%by its start and its end point. It is organized as [xstart ystart zstart; xend
%yend zend]. To plot a vector in this form, the function plotvector() has been
%implemented. To first plot our coordinate system plotaxes() can be used, which
%in turn calls plotvector() three times. The arguments of plotaxes are explained
%later on.

clear
hold( "off" )
rot = eye( 3 );                  %see below for details
plotaxes( rot, 3, 0.05, 0, "k" ) %k=black
hold( "on" )                     %following plot commands will add to the former output
vector = [ 0, 0, 0; 1 1 1 ];
plotvector( vector, "r" )        %r=red
title( "Plot Vector" );
kbhit();                         %press return to continue the script


%Tilt compensation of a compass is dealing heavily with rotations. A rotation in
%3D can be described with a 3x3 rotation matrix. Every rotation can be split up
%into the three rotations about the axes of the coordinate system. To easily
%define a rotation about the x-, y- and z-axis respectively the three functions
%rotx(), roty() and rotz() have been implemented. The total rotation is then
%simply the multiplication of all three.

%The function plotaxes() lets us define the attitude of the coordinate system to
%draw. The first parameter is the rotation, the second the length of each axis.
%The next 3 parameters define the size/position of the XYZ characters drawn at
%the axes and the color. Type 'help plotaxes' for details.
%Lets again draw our black reference system. The rotation is 0, so we call
%plotaxes() with the unity matrix eye(3) as rotation. Then lets draw another
%blue coordinate system with some arbitrary attitude. Don't be bothered by the
%angles, direction of rotations etc. We'll get to that later.

clear
hold( "off" )
rot = eye( 3 );
plotaxes( rot, 3, 0.05, 0, "k" )
hold( "on" )
rot = rotx( 10 ) * roty( 10 ) * rotz( 30 );
plotaxes( rot, 1, 0.05, 0, "b" )
title( "Plot Coordinate Systems" );
kbhit();

%Octave lets us rotate a vector given in the form above simply by multiplying it
%with the rotation matrix, thus vector_rotated = vector * rot. This will rotate
%both, the start and the end point, at the same time.

vector = [ 0, 0, 0; 1 1 1 ];
plotvector( vector, "r" )
vector_rotated = vector * rot;
plotvector( vector_rotated, "m" )
title( "Plot Rotated Vector" );
kbhit();

%Ok, now let's go into that a bit more precisely. Like the definition of the
%coordinate system (relativ orientation of the axes, where is positive etc.) it
%is also required to define the direction of rotations. We define the rotations
%mathematicaly positive. Seen the xy-plane from above (looking in the negative
%z direction), the rotation about z is counter-clockwise. The x-axis is 0
%degrees, the y-axis 90 degrees. The same for the other axes.

draw_definition
title( "Definition of Axes and Angles" );
kbhit();

%It is essential to understand that in a sequence of rotations the rotations
%can't be exchanged (non-commutative). In the following the result is not the
%same:

clear
hold( "off" )
rot = eye( 3 );
plotaxes( rot, 3, 0.05, 0, "k" )
hold( "on" )
rx = rotx( 20 );
ry = roty( 30 );
vector = [ 0, 0, 0; 1 1 1 ];
plotvector( vector * rx * ry, "r" ) %first rx, then ry
plotvector( vector * ry * rx, "r" ) %the other way round
title( "Non-Commutative Rotations" );
kbhit();

%Thus we need to define which angle comes first. Intuitively the sequence is x,
%then y, then z. We will see later in the calculations why this is a clever
%choice.
%To reverse a rotation the angle is negated and the sequence is exchanged.

clear
hold( "off" )
rot = eye( 3 );
plotaxes( rot, 3, 0.05, 0, "k" )
hold( "on" )
vector = [ 0, 0, 0; 1 1 1 ];
vector_rot     = vector     * rotx(  20 ) * roty(  30 );
vector_rotback = vector_rot * roty( -30 ) * rotx( -20 )
plotvector( vector, "r" )
plotvector( vector_rot, "m" )
plotvector( vector_rotback, "k" )
title( "Reversal of Rotations" );

%Furthermore we define that a vehicle or an airplane is heading in the positive
%x direction (if it is not rotated). The wings of an airplane would be level
%with the xy-plane, its longitudinal axis on the x-axis. In this initial
%position all angles are 0. The attitude, the deviation from the initial
%position, can now be described by three angles:
% roll  about the x-axis
% pitch about the y-axis
% yaw   about the z-axis

%It doesn't matter if an angle is given as 0..360 degrees or -180..+180 degrees.
%For example -90 is equal to +270 degrees. The trigonometric functions of Octave
%used later on output -180 to +180 degrees. So I prefer to use consistently
%-180..+180 degrees.

%There is still an ambiguity left. The same attitude can be expressed by
%different combinations of roll, pitch and yaw.
%Turning the vehicle upside down can be expressed by either pitch > +-90 or
%roll > +-90 degrees. For example a pitch of 160 degrees is the same as rolling
%180 degrees, then applying a pitch of 20 degrees and flipping the vehicle in
%the other direction. In both cases the attitude and the rotation matrix is the
%same:
% roll  =   0       roll  = 180
% pitch = 160  <=>  pitch =  20
% yaw   =   0       yaw   = 180

%When applying a pitch larger than +-90 degrees the nose of the vehicle is
%suddenly pointing backwards. This changes the heading without altering yaw.
%Also the direction of roll is reversed. An ill-suited behaviour.
%Roll, on the other hand, does not alter the heading. Even when turning the
%vehicle upside down with roll larger than +-90 degrees.
%Therefore, by convention, pitch is limited while the full range for roll is
%admitted:
% -180 <  roll  <= 180
%  -90 <= pitch <=  90
% -180 <  yaw   <= 180

%(To be exact: The combinations 0/+-90/0 and 180/+-90/180 describe the same
%attitude. Nevertheless they are both admitted. Why? +-90 is a special case
%anyway. Heading is not defined. Roll (and with that all other angles) can't be
%determined from the sensor outputs. We will see that later. So +-90 degrees has
%to be excluded from the final heading calculation anyway. Admitting both
%combinations allows to decide only on pitch during the conversion of
%combinations.)

%By looking at some combinations describing the same attitude we can derive a
%method to convert between them. Converting an angle combination not conforming
%to the convention is done by flipping all directions and inverting pitch:

clear
disp( "Rotation matrices for equal angle combinations:" )
roll  =   0
pitch = 160
yaw   =   0
rot = rotx( roll ) * roty( pitch ) * rotz( yaw )
if( (pitch > 90) || (pitch < -90) )
   roll = roll + 180;  %flip
   if( roll > 180 )    %pull into standard range
      roll -= 360;
   endif
                  
   pitch = pitch + 180;
   if( pitch > 180 )
      pitch -= 360;
   endif
   pitch = -pitch;     %invert pitch
                  
   yaw = yaw + 180;
   if( yaw > 180 )
      yaw -= 360;
   endif
endif
roll
pitch
yaw
rot = rotx( roll ) * roty( pitch ) * rotz( yaw )

%Later on in the next parts of this tutorial we will do both, calculating sensor
%outputs from angles given and - the other way round - calculating angles from
%sensor outputs.
%You can go into the calculations with any arbitrary angle combination. The
%angles calculated, however, will conform to the convention above. This is
%analog to the behaviour of trigonometric functions. The ranges of the inverse
%functions are subsets of the domains of the original functions.

%With all these conventions the attitude, the deviation from the initial position,
%can now unambiguously be described.

%Now that we are all setup we can go to the actual math. Tilt compensation is
%done in two steps. First the attitude of the vehicle or airplane is determined
%using the output of an acceleration sensor. This is described in part 3 of this
%tutorial. The second step is the heading calculation including the tilt
%compensation based on the attitude. Part 4 will go into details about this.
