%part3.m
% -- Script File: part3
%     The files part1.m to part4.m are both, the text of a tutorial as well as
%     the tutorials Octave scripts. You may just read through them though it is
%     quite handy to see the math actually executed. Walk your way through the
%     parts. Part 1 starts with the Octave setup.
%
%     Arnold Neugebauer

page_screen_output( 0 )     %don't use pager
subplot( 1, 1, 1 )          %set to standard single plot mode

%PART 3
%======

%Determining the Attitude
%------------------------
%In the following we will derive a model. That is we will draw the ground-fixed
%reference system, the tilted vehicle-fixed system and acceleration vectors.
%This is quite handy. We can play with the model and see what happens. Finally,
%by expanding the calculation with matrices and vectors, we get the equations to
%implement in the electronic compass.

%The following defines an arbitrary attitude and plots the ground-fixed as well
%as the vehicle-fixed system. Also the positive and negative rotation matrices
%are calculated beforehand.

clear
hold( "off" )
plotaxes( eye(3), 30, 0.5, 0, "k" )
hold( "on" )

roll  =  30; %x
pitch =  10; %y
yaw   =  10; %z

rxp = rotx(  roll );
ryp = roty(  pitch );
rzp = rotz(  yaw );
rxn = rotx( -roll );
ryn = roty( -pitch );
rzn = rotz( -yaw );
rp  = rxp * ryp * rzp; %first rotate about x, then y, then z
rn  = rzn * ryn * rxn; %sequence the other way round!

plotaxes( rp, 10, 0.5, 0, "b" )

%Next the acceleration vector. It points up. We just make everything in the
%model positive for now. It's easier to understand. Later on we can make it
%point down just by exchanging signs.
%The magnitude is arbitrarily chosen. In a real system the magnitude is the
%maximum sensor output in the earth field found out by calibration.
%The postfix _g means that a vector is described based on the ground-fixed
%coordinate system. Variables with postfix _v are described based on the
%vehicle-fixed coordinate system.

acc_strength = 18;
acc_g = [ 0 0 0; 0 0 acc_strength ];
plotvector( acc_g, "r" )
title( "Initial Position" )
kbhit();

%At this point we have our reference system and a vehicle system somehow tilted
%in the 3D space. acc_g is an object in the space. For calculations in the
%vehicle system, the general trick is to transform the tilted vehicle system to
%the reference system. Calculations are done conveniently there in a standard
%way. All objects are transformed in the same way, so that their relative
%position to the vehicle system remains constant. After the calculations have
%been done the vehicle system and all objects are transformed back / tilted
%again. The transformations are coordinate transformation or rotations
%respectively. Calculations here are just the simple split-up of a vector in its
%three components.

%The next Octave commands draw the vehicle-fixed system and the acceleration
%vector again. This time rotated against the attitude of the vehicle. The
%vehicle-fixed system becomes the same as the ground-fixed system. The
%acceleration vector is now slanted but its position relative to the
%vehicle-fixed system has not changed.

hold( "off" )
plotaxes( eye(3), 30, 0.5, 0, "k" )
hold( "on" )
plotaxes( rp * rn, 10, 0.5, 0, "b" ) %same as plotaxes( eye(3), 10, 0, "b" )
acc_v = acc_g * rn;
plotvector( acc_v, "r" )
title( "Rotated Against Attitude" )
kbhit();

%Now the accelertion vector can be split up into its components parallel to the
%axes. The magnitude of these vectors is the output of the acceleration sensors
%experienced. The three vectors along the axes are:

accx_v = [ 0, 0, 0; acc_v(2, 1), 0, 0 ];
accy_v = [ 0, 0, 0; 0, acc_v(2, 2), 0 ];
accz_v = [ 0, 0, 0; 0, 0, acc_v(2, 3) ];
plotvector( accx_v, "g" )
plotvector( accy_v, "g" )
plotvector( accz_v, "g" )
title( "Acceleration Vector Components" )
kbhit();

%Rotate components back to their original attitude:

hold( "off" )
plotaxes( eye(3), 30, 0.5, 0, "k" )
hold( "on" )
plotaxes( rp, 10, 0.5, 0, "b" )
plotvector( acc_g, "r" )
accx_g = accx_v * rp;
accy_g = accy_v * rp;
accz_g = accz_v * rp;
plotvector( accx_g, "g" )
plotvector( accy_g, "g" )
plotvector( accz_g, "g" )
title( "Rotated Back to Initial Position" )
kbhit();

%To see more precisely that the green components on the axes are in fact adding
%up to the red acceleration vector this can be drawn a little bit different.
%Let's graphically add up the green vectors of the vehicle-fixed system. accx_g
%starting from zero, accy_g vector starting at end of accx_g, accz_g vector
%starting from there. The end point of these 3 vectors must be the end point of
%acc_g, every single vector pointing in the direction of the respective
%vehicle-fixed axis.

accy_g_offs = [ accx_g(2,:); accx_g(2,:) + accy_g(2,:) ];
accz_g_offs = [ accx_g(2,:) + accy_g(2,:); accx_g(2,:) + accy_g(2,:) + accz_g(2,:) ];
hold( "off" )
plotaxes( eye(3), 30, 0.5, 0, "k" )
hold( "on" )
plotaxes( rp, 10, 0.5, 0, "b" )
plotvector( acc_g, "r" )
plotvector( accx_g, "g" )
plotvector( accy_g_offs, "g" )
plotvector( accz_g_offs, "g" )
title( "Sum of Acceleration Vector Components" )
kbhit();

%What we are finally interested in is the relationship between the acceleration
%sensor outputs (the magnitude of the components seen on the vehicle-fixed axes)
%and the attitude of the vehicle.
%The sensor outputs are acc_v(2,:), forming the vector acc_v in the
%vehicle-fixed coordinate system. For convenience these abbreviations are used
%in the following:

accx = acc_v(2,1);
accy = acc_v(2,2);
accz = acc_v(2,3);

%Expanding the equation acc_v = acc_g * rn from above leads to the final
%equations. Now we see why it was clever to choose the rotation sequence xyz for
%the vehicle. Rotating the other way round is zyx. Since acc_g is pointing from
%the origin upwards it is invariant to the first rotation about z. So we can
%leave that away, simplifying the equations enormously.
%   acc_v = acc_g * rn
%=> [ accx, accy, accz ] = [0 0 acc_strength ] * ryn * rxn
%=> accx =   acc_strength * sind( pitch )                 (1)
%   accy = - acc_strength * sind( roll ) * cosd( pitch )  (2)
%   accz =   acc_strength * cosd( roll ) * cosd( pitch )  (3)
%There are more equations than unknown variables what allows to solve in
%different ways. Simply taking (1) and (2) leads to the equations often found
%in the literature. The sign must be changed since in rn the angles are
%negative.

disp( "Roll and pitch calculated using arc sine:" )
pitch1 = - asind( accx / acc_strength );               %y
roll1  =   asind( accy / acc_strength / cosd(pitch1) ) %x
pitch1
kbhit();

%However, doing this in the real world is disadvantageous. The slope of the asin
%function is strongly increasing with higher angles. Thus it gets very sensitive
%to noise, offset and gain errors. The following shows this by adding 10% noise
%to the sensor outputs (the high noise level is arbitrarily chosen but seems not
%to be unusual in a low-quality/low-cost circuit).

draw_asin_error()
kbhit();

%The output can be improved significantly. Dividing equation (1) by (3) and
%equation (2) by (3) leads to expressions with atan. The slope of the atan
%function is increasing later than the asin function.
%Note that acc_strength is canceled down. So we don't need to know the absolute
%value from now on.

disp( "Roll and pitch calculated using arc tangent:" )
roll1  =   atand( accy / accz )               %x
pitch1 = - atand( accx / accz * cosd(roll1) ) %y
kbhit();

draw_atan_error()
kbhit();

%Still, the error with pitch near 90 degrees gets very high. Limiting the
%operation to about +-80 degrees is acceptable for vehicles or handheld devices.
%Otherwise, however, more effort is needed to get around the noise near the
%poles of the trigonometric functions.
%In any case, when dealing with a real system, proper noise filtering/damping of
%the sensor outputs has to be applied and a plausibility check to cancel out the
%acceleration of the vehicle itself.

%Remember the conventions made for the output ranges in part 2 of this tutorial?
%We targeted for -180 < roll <= 180 and -90 <= pitch <=  90. A pitch of exactly
%+-90 degrees, however, leads to accy = accz = 0. Roll (and with that all other
%angles) can't be determined from the sensor outputs. So a pitch of +-90 degrees
%has to be excluded from the calculations (by discarding measurements with
%accy = accz = 0). Heading is not defined anyway in this case.

%The last equations for roll and pitch are using atan with an output range of
%-90 < angle < 90. While this is perfect for pitch, we need to use atan2 for
%roll. atan2 is a variant of the atan function with the expanded output range
%of -180 < angle <= 180. For details type 'help atan2' on the Octave command
%line.

%Since atan2 is part of the C programming language's math.h standard library, it
%is often available. Also Octave features the built-in function. atan2d.m shows
%a possible implementation if it is not available on your system. (Be aware that
%sometimes implementations reverse the order of the parameters.)

%Alter roll, pitch and yaw at the beginning of the script to see the effect:

disp( "Roll and pitch calculated using arc tangent 2:" )
roll1  =   atan2d( accy, accz )               %x
pitch1 = - atand( accx / accz * cosd(roll1) ) %y

%In this part we derived a handy model to play around with and to verify our
%calculations. It will also be used in the next part. And we derived two
%equations, enabling us to determine the tilt of the vehicle-fixed system from
%its acceleration sensor outputs. So let's start with the actual heading
%calculation in the next part.
