%part4.m
% -- Script File: part4
%     The files part1.m to part4.m are both, the text of a tutorial as well as
%     the tutorials Octave scripts. You may just read through them though it is
%     quite handy to see the math actually executed. Walk your way through the
%     parts. Part 1 starts with the Octave setup.
%
%     Arnold Neugebauer

page_screen_output( 0 )     %don't use pager
subplot( 1, 1, 1 )          %set to standard single plot mode

%PART 4
%======

%Heading Calculation
%-------------------
%This part deals with the heading calculation. We will use the same approach as
%in the last part. First the whole flow is modeled by manipulating vectors and
%using Octave's matrix multiplications. Intermediate steps and results are
%visualized . Remember that you can drag the graphics around with the mouse. And
%be encouraged to alter the parameters and calculations. Play around with it to
%get a feeling of what's going on.
%In a second step, by expanding the calculation with matrices and vectors, we
%get the equations to implement in the electronic compass.

%Again, the following defines an arbitrary attitude and plots the ground-fixed
%as well as the vehicle-fixed system. Also the positive and negative rotation
%matrices are calculated beforehand.

clear
hold( "off" )
plotaxes( eye(3), 30, 0.5, 0, "k" )
hold( "on" )

roll  = 30; %x
pitch = 10; %y
yaw   = 10; %z

rxp = rotx(  roll );
ryp = roty(  pitch );
rzp = rotz(  yaw );
rxn = rotx( -roll );
ryn = roty( -pitch );
rzn = rotz( -yaw );
rp  = rxp * ryp * rzp; %first rotate about x, then y, then z
rn  = rzn * ryn * rxn; %sequence the other way round!

plotaxes( rp, 10, 0.5, 0, "b" )

%Now the magnetic field vector. It points up. Like the acceleration vector we
%just make everything positive for now. (Later, pointing down with an
%inclination = 70 degrees is more appropriate.)

%As of our definition from part 2, north is in the direction of the x-axis of
%the ground-fixed system. Given the parameters strength and inclination the
%following models the magnetic field vector in the xz-plane. The magnetic
%strength has been arbitrarily chosen to fit in our coordinate system drawn. In
%a real system the strength is the maximum sensor output in the earth field
%found out by calibration. A magnetic field vector with a negative inclination
%is pointing up (towards +z) per definition.

mfield_strength = 18;
mfield_inclination = -70;        %northern europe would be about +70 degrees
mfieldx_g = mfield_strength * cosd( mfield_inclination );
mfieldy_g = 0;
mfieldz_g = sqrt( mfield_strength^2 - mfieldx_g^2 );   %derived from vector length
if( mfield_inclination > 0 )
   mfieldz_g = -mfieldz_g;
endif
mfield_g  = [ 0, 0, 0; mfieldx_g, mfieldy_g, mfieldz_g ];
plotvector( mfield_g, "r" )
title( "Initial Position" )
kbhit();

%The question is now what sensor outputs we see on the axes of the vehicle-fixed
%system. The same trick as for the acceleration in part 3 is used. The tilted
%vehicle system and all objects are transformed to the ground-fixed reference
%system. The relative position is kept. Then calcutions are executed and finally
%everything is tranformed back.

%The next commands draw the vehicle-fixed system and the magnetic field vector
%again. This time rotated in the opposite direction of the vehicle attitude. The
%vehicle-fixed system becomes the same as the ground-fixed system. The magnetic
%field vector is not in the xz-plane anymore but its position relative to the
%vehicle-fixed system has not changed.

hold( "off" )
plotaxes( eye(3), 30, 0.5, 0, "k" )
hold( "on" )
plotaxes( rp * rn, 10, 0.5, 0, "b" ) %same as plotaxes( eye(3), 10, 0.5, 0, "b" )
magn_v = mfield_g * rn;
plotvector( magn_v, "r" )
title( "Rotated Against Attitude" )
kbhit();

%Now the magnetic field vector can be split up into its components parallel to
%the axes. The magnitude of these vectors is the output of the magnetic field
%sensors experienced. The three vectors along the axes are:

magnx_v = [ 0, 0, 0; magn_v(2, 1), 0, 0 ];
magny_v = [ 0, 0, 0; 0, magn_v(2, 2), 0 ];
magnz_v = [ 0, 0, 0; 0, 0, magn_v(2, 3) ];
plotvector( magnx_v, "g" )
plotvector( magny_v, "g" )
plotvector( magnz_v, "g" )
title( "Magnetic Field Vector Components" )
kbhit();

%Rotate components back to their original attitude:

hold( "off" )
plotaxes( eye(3), 30, 0.5, 0, "k" )
hold( "on" )
plotaxes( rp, 10, 0.5, 0, "b" )
plotvector( mfield_g, "r" )
magnx_g = magnx_v * rp;
magny_g = magny_v * rp;
magnz_g = magnz_v * rp;
plotvector( magnx_g, "g" )
plotvector( magny_g, "g" )
plotvector( magnz_g, "g" )
title( "Rotated Back to Initial Position" )
kbhit();

%Graphically add up the green vectors to see more precisely that they are in
%fact add up to the red magnetic field vector:

magny_g_offs = [ magnx_g(2,:); magnx_g(2,:) + magny_g(2,:) ];
magnz_g_offs = [ magnx_g(2,:) + magny_g(2,:); magnx_g(2,:) + magny_g(2,:) + magnz_g(2,:) ];
hold( "off" )
plotaxes( eye(3), 30, 0.5, 0, "k" )
hold( "on" )
plotaxes( rp, 10, 0.5, 0, "b" )
plotvector( mfield_g, "r" )
plotvector( magnx_g, "g" )
plotvector( magny_g_offs, "g" )
plotvector( magnz_g_offs, "g" )
title( "Sum of Magnetic Field Vector Components" )
kbhit();

%Ok - after having the model up and running we can now try to figure out the
%equations for the implementation. What we are finally interested in is the
%relationship between the magnetic field vector components seen in the
%vehicle-fixed system (the sensor outputs) and the heading.
%In theory it would be possible to derive them directly from the model above
%(like we did for the attitude in part3). We would have to expand
%magn_v = mfield_g * rn. Unfortunately math with the full rotation rn is very
%hard to do. (For the attitude we were lucky - the acceleration vector only had
%a z-component and the rotation sequence had been chosen to support further.)

%What helps is an intermediate step: First calculate the tilt-compensated sensor
%outputs and then do the final heading calculation without tilt in the xy-plane.
%This way the rotation about z is omitted in the first step to make the
%expansion easier. Then the z-component is left out since it does not contain
%any heading information, further simplifying things. In the following you will
%see what I mean.
%For a better understanding let's draw the vehicle-fixed system in its
%tilt-compensated position. It is rotated only about z with rzp. The xy-plane is
%level with the xy-plane of the ground-fixed system. Here again we see why it
%was clever to choose the rotation sequence xyz for the vehicle. Only with z
%beeing the last roation, the yaw is equal to the heading and the rotations
%about x and y can simply be left out.

hold( "off" )
plotaxes( eye(3), 30, 0.5, 0, "k" )
hold( "on" )
plotaxes( rzp, 10, 0.5, 0, "b" )
plotvector( mfield_g, "r" )
title( "Tilt Compensated Position" )
kbhit();

%To calculate the magnetic field components we would see on the axes of the
%tilt-compensated system, the trick is again to do a transformation to the
%ground-fixed system. This time only rzn is needed, not the full rn.
%(I skip rotating back for display from now on, otherwise its the same game as
%always - rotating, split up, rotating back.)

hold( "off" )
plotaxes( eye(3), 30, 0.5, 0, "k" )
hold( "on" )
plotaxes( rzp * rzn, 10, 0.5, 0, "b" ) %same as plotaxes( eye(3), 10, 0.5, 0, "b" )
plotvector( mfield_g * rzn, "m" )
magnc_v = mfield_g * rzn;                    %equation (1)
magnxc_v = [ 0, 0, 0; magnc_v(2, 1), 0, 0 ];
magnyc_v = [ 0, 0, 0; 0, magnc_v(2, 2), 0 ];
magnzc_v = [ 0, 0, 0; 0, 0, magnc_v(2, 3) ];
magnyc_v_offs = [ magnxc_v(2,:); magnxc_v(2,:) + magnyc_v(2,:) ];
magnzc_v_offs = [ magnxc_v(2,:) + magnyc_v(2,:); magnxc_v(2,:) + magnyc_v(2,:) + magnzc_v(2,:) ];
plotvector( magnxc_v, "g" )
plotvector( magnyc_v_offs, "g" )
plotvector( magnzc_v_offs, "g" )
title( "Tilt-Compensated Magnetic Field Vector Components" )
kbhit();

%Note that the variables got an additional 'c' signifying the tilt-compensation.
%The magnetic field vector magnc_v seen in the tilt-compensated position is
%drawn in magenta. Keep in mind equation (1). It is needed later on.

%In our model we already did the transformation from the vehicle-fixed to the
%ground-fixed system. For reference the magnetic field vector magn_v - the
%vector without tilt-compensation - is added again to this drawing:

%magn_v = mfield_g * rn   %equation (2), already calculated above
plotvector( magn_v, "r" )
magny_v_offs = [ magnx_v(2,:); magnx_v(2,:) + magny_v(2,:) ];
magnz_v_offs = [ magnx_v(2,:) + magny_v(2,:); magnx_v(2,:) + magny_v(2,:) + magnz_v(2,:) ];
plotvector( magnx_v, "g" )
plotvector( magny_v_offs, "g" )
plotvector( magnz_v_offs, "g" )
title( "...+ Magnetic Field Vector Components Without Tilt" )
kbhit();

%So, how do we get from the experienced magnetic field (the senor outputs, red)
%to the tilt-compensated magnetic field (magenta)?
%By eliminating the unknown mfield_g from equation (1) and (2) we get the
%relation between them:
%    magnc_v = mfield_g * rzn        equation (1) from above
%    magn_v  = mfield_g * rn         equation (2) from above with rn = rzn * ryn * rxn
% => magn_v  = magnc_v  * ryn * rxn
%<=> magnc_v = magn_v   * rxp * ryp  equation (3)

%The last two equations may have been simply developed from the idea that it is
%possible to convert back and forth just by leaving out the rotation about z.
%(Though, I think, it is not obvious that this is also mathematically correct.
%And in fact it wouldn't be that simple if we hadn't chosen the rotation
%sequence to be xyz.)

%By expanding equation (3) we get the following to finally calculate the
%heading:
magnx  = magn_v(2,1); %abbreviation for convenience
magny  = magn_v(2,2);
magnz  = magn_v(2,3);
magnxc = magnx * cosd( pitch ) + magny * sind( roll ) * sind( pitch ) + magnz * cosd( roll ) * sind( pitch );
magnyc = magny * cosd( roll )  - magnz * sind( roll );

%The z-component of magnc_v naturally is not contributing to the heading. The
%heading can finally be determined from the arrangement in the xy-plane. To
%better see this look at the drawing from the top:

hold( "off" )
plotaxes( eye(3), 30, 0.5, 1, "k" )
hold( "on" )
plotaxes( eye(3), 10, 0.5, 1, "b" )
plotvector( mfield_g * rzn, "m" )
plotvector( magnxc_v, "g" )
plotvector( magnyc_v_offs, "g" )
plotvector( magnzc_v_offs, "g" )
handle_to_current_axis_object = gca();
set( handle_to_current_axis_object, "view", [0, 90] )
title( "Tilt-Compensated Magnetic Field Vector Components" )
kbhit();

%From the figure we derive the equation:
%heading = - atand( magnyc / magnxc )

%As with roll atan2 has to be used for an expanded output range. To stay well in
%the range of -180 < heading <= 180, change of sign is done by changing the sign
%of the first parameter:
heading = atan2d( - magnyc, magnxc )

%We can now check if our equations work for every attitude as expected. The
%stepsize in the script is currently set to 15 degrees to save time. Reduce it
%if you like:
angle_check


%I hope you are still with me ;-) At the end the whole stuff boils down to only
%5 equations. In recapitulation here is an abstract with some additional
%implementation hints. The processing sequence how I think it should be
%implemented is:
%  3 axis acceleration measurement (accx, accy, accz)
%  3 axis magnetic field measurement (magnx, magny, magnz)
%  noise filter / damping
%  offset and gain error compensation
%  discard measurements with accz = 0
%  plausibility check, sqrt( accx^2 + accy^2 + accz^2 ) close to 1g
%  roll  =   atan2d( accy, accz )
%  pitch = - atand( accx / accz * cosd(roll) )
%  (limit operational range to +-80 degrees)
%  magnxc  = magnx * cosd( pitch ) + magny * sind( roll ) * sind( pitch ) + magnz * cosd( roll ) * sind( pitch )
%  magnyc  = magny * cosd( roll )  - magnz * sind( roll )
%  heading = atan2d( - magnyc, magnxc )

%The 6 measurements must be executed in fast succession, virtually at the same
%time. This makes up a snapshot of the current attitude and magnetic field. All
%measurements are belonging together.

%The error propagation of the equations allows to do the noise filtering either
%directly on the sensor outputs or later on intermediate results or even on the
%final heading. The result is about the same.
%The noise filtering is typically done on the sensor outputs. The reason is that
%the sensor outputs are usually sampled quite fast (faster than the change in
%attitude or at least tracking it very fast) while the display requires only a
%slow update rate. With the filter on the sensor outputs, the performance
%critical trigonometric calculations are done only with the dispay update rate.

%The noise filter is primarily intended to cancel out the noise. This could be
%done by simple averaging. With damping I mean the more tricky part to achieve
%that the displayed heading nicely follows the compass/vehicle movement (like a
%liquid damped mechanical compass). Damping can be combined with the noise
%filter or implemented later, working on the final heading for example.

%Offset and gain error compensation is done with the values from the compass
%calibration. In its simplest form (an offset and a factor for each axis) it
%makes up for the linear and time-invariant deviations of the hardware.

%Applying offsets and scaling each axis separately provides also a simple way to
%calibrate for soft and hard iron effects of the compass environment. This
%method is sufficient in a lot of environments (environments producing the
%typical Lissajous-formed deviation in the xy-plane).

%To avoid division by 0 measurements with accz = 0 are simply discarded.
%accz = 0 means roll and/or pitch = +-90 degrees. This is above the limit we
%considered to be useful for vehicles/handhelds. pitch = +- 90 degrees leads to
%accy = accz = 0 and thus roll can't be determined from sensor outputs. On one
%hand it's pretty useless anyway. In practice getting exactly 0 from the sensors
%happens quite seldom. So think of it as a measure to get the software
%bulletproof.

%Acceleration of the vehicle itself will lead to a false tilt detection. It is
%assumed that the vehicle acceleration is small and/or from short duration. The
%plausibility check avoids false tilt detection. If the magnitude of the
%acceleration measured is not close to 1g, the current measurement is discarded.
%It is not entering the noise filter.

%As discussed in part 3, noise sensitivity is very high near the poles of the
%trigonometric functions. Limiting the operation to about +-80 degrees is
%acceptable for vehicles/handhelds. Otherwise, however, more effort is needed to
%get around the noise near the poles of the trigonometric functions.

%All the stuff discussed for a possible implementation taking various error
%sources into account is simulated by the script
hold( "off" )
implementation



%I hope this was not too scary. I tried to shed light on every corner of the
%tilt compensation, and to provide you with simulation models for your own
%experiments. Hopefully I got you to a point where you are now able to work
%successfully on your projects.
%Good Luck :-)
